/*
 *  NAME
 *	libnew.c:
 *	    mnew, cnew, renew, odelete - replacing malloc, calloc, realloc, free
 *	    odeleted - gives a little memory statistic at the end
 * 
 *  SYNOPSIS
 *	include <onew.h>
 * 
 *	void *malloc (size_t size)
 *	void *calloc (size_t nobj, size_t size)
 *	void *realloc (void *p, size_t size)
 *	void free (void *p)
 *	void freed ()
 *	void allog_begin (const char *logname)
 *	void allog_end ()
 * 
 *  DESCRIPTION
 *	Replaces the original malloc, calloc and realloc function for easier
 *	debugging (see below).
 * 
 *	At the end of the programm the function freed must be called to get a
 *	statistic if more memory is allocated than given free.
 * 
 *  COMMON ERRORS FOR DYNAMICAL MEMORY ALLOCATION
 *	(1) missing test of the result of 'new'
 *	(2) required memory allocation was to small and the block limits
 *	    are overwritten
 *	(3) block not given back (dead memory body)
 *	(4) invalid pointer given to 'delete' (multiple deallocating,
 *	    wrong pointer)
 *	(5) assumption that the block is cleared (with zeros)
 *	(6) access to a freed block
 *	(7) memory reallocated and moved to a new allocation but there are
 *	    still some pointers which points to the old location
 * 
 *  SOLUTION
 *	(2) block end is marked with a magic word
 *	(3) manages a list of blocks; gives at the end of the program a list of
 *	    all still reserved blocks
 *	(4) a) compares the pointer with the block list
 *	    b) marks the beginning of the block with a magic word
 *	(5) fill the block with non-zero values
 *	(6) like (5)
 *	(7) return ALWAYS a new location and shredders old one (like (5))
 * 
 *  DIAGNOSTICS
 *	"source.c(#): WARNING: 0 bytes ordered!"
 *	"source.c(#): arithmetik overflow in 'calloc (#, #)'"
 *	"source.c(#): NULL pointer to delete!"
 *	"source.c(#): Invalid pointer to delete!"
 *	"source.c(#): End of #-byte-block - reserved at source.c(#) -
 *		      overwritten!"
 *	"source.c(#): Can't reallocate NULL-pointer!"
 *	"source.c(#): Invalid pointer to reallocate!"
 *	"source.c(#): Invalid pointer!"
 *	"source.c(#): # blocks reserved, but only # freed!"
 *	"source.c(#): # bytes reserved at source.c(#) not freed!"
 *	"source.c(#): WARNING: strdup() called with NULL pointer!"
 *	"source.c(#): The heap is corrupted - trying to restore..."
 * 
 *  BUGS
 *	cnew() is not very efficient because it calls mnew() and clears then
 *	allocated memory with 0.
 *
 *	The restoration of a corrupted heap is not tested.
 *	
 *	Overhead of 32 Bytes for each memory allocation.
 * 
 *  VERSION $Revision: 3.3 $ ($Date: 2000/02/27 15:49:36 $)
 *
 *  COPYRIGHT (c) 1991-2000 by O. Boehm (boehm@ba-stuttgart.de)
 *	Permission to use, copy, modify, and to distribute this software and its
 *	documentation for any purpose is hereby granted without fee, provided
 *	that the above copyright notice appear in all copies and that both that
 *	copyright notice and this permission notice appear in supporting
 *	documentation. There is no representations about the suitability of
 *	this software for any purpose. This software is provided "as is" without
 *	express or implied warranty.
 * 
 *  HISTORY
 *	created:
 *	    20-Nov-91 by coreproducer O. Boehm
 *	modified:
 *	    12-Jan-93 by coreproducer O. Boehm
 *		- file name and line number added to error message
 *		- adapted to ANSI-C
 *	    19-Apr-93 coreproducer O. Boehm
 *		- malloc(0), calloc(..0..) now produces a warning!
 *		- adapted for ANSI-C AND K&R-C
 *	    18-Jul-94 by coreproducer O. Boehm
 *		- <malloc.h> included now
 *		- mdelete(d) changed to odelete(d) because of conflicts with a
 *		  ftp library
 *		- support of strdup()
 *	    May-95 by coreproducer O. Boehm
 *		- each realloc() call returns a new pointer and shredders the
 *		  old area
 *		- Block struct extended by file name and line
 *		- reserved blocks managed in a list now
 *		- heap is checked for each allocation and free'ing
 *	    May-95 (cont'd)
 *		- using LWP library for heap checking on a Sun
 *	    Jul-95  by coreproducer O. Boehm
 *		- name changed to libonew.c to avoid conflicts with other libs
 *		- allog_begin(), allog_end() added
 *	    Mar-96 by coreproducer O. Boehm
 *		- adaption to Solaris 2.5
 *	    Oct-96 by coreproducer O. Boehm
 *		- adaption to RCS
 *	    Feb-2000 by Oliver Boehm
 *		- adaption to Linux 2.2
 *
 *	$Log: libonew.c,v $
 *	Revision 3.3  2000/02/27 15:49:36  oliver
 *	adapted to linux 2.2
 *
 *	Revision 3.2  1996/10/24 08:40:40  lve163d
 *	adaption to RCS
 *
 *  SEE ALSO
 *	Programmieren in C++
 *	Kursbegleitendes Material
 *	Sun Microsystems GmbH (Ausbildungszentrum)
 *	p. 123 - 126
 */



#ifndef lint
static const char ident[]="$Id: libonew.c,v 3.3 2000/02/27 15:49:36 oliver Exp $ Copyr O. Boehm";
#endif



#if defined(__STDC__) || defined (__cplusplus)
#include <stdarg.h>
#else
#include <varargs.h>
#endif
#include <time.h>
#include "libonew.h"





#ifndef MIN
#define MIN(x,y)	((x) < (y)) ? (x) : (y)
#define MAX(x,y)	((x) < (y)) ? (y) : (x)
#endif

#if defined(sparc)			/* only 4-Byte adresses are valid */
#define VALID_BLOCK_POINTER(ptr)	(((unsigned long)(ptr) % 4) == 0)
#else
#define VALID_BLOCK_POINTER(ptr)	(1 == 1)	       /* always true */
#endif


/*  return codes  */

#define NO_ERROR			0
#define ERR_INVALID_POINTER		1
#define ERR_MEMORY_OVERWRITTEN		2

#define HEAP_OK				0
#define HEAP_CORRUPTED			-1





static const long beg_magic    = 0x12345678;
static const char end_magic[4] = {'B', '', 'h', 'm'};



typedef struct {
    int allocated;
    int freed;
} Counter;

typedef struct s_Block {
    struct s_Block  *prev;		/* previous allocated block */
    struct s_Block  *next;		/* next allocated block */
    size_t	    size;		/* ordered size */
    char	    *file;		/* __FILE__ */
    int		    line;		/* __LINE__ */ 
    long	    magic;		/* begin pattern */
    char	    *p_block_end;	/* end pattern (can be stored */
    					/* also to an odd adress)     */
} Block;



static Block *p_FirstBlock = NULL;	/* pointer to the first and    */
static Block *p_LastBlock = NULL;	/* last allocated memory block */

static Counter BlockCounter = {0, 0};





/* --------   P R O T O T Y P E S   ( see -> onew.h )  ---------------------- */

void *mnew _((size_t size, char* file, int line));
void *cnew _((size_t nobj, size_t size, char* file, int line));
void *renew _((void *p, size_t size, char* file, int line));
void odelete _((void *p, char* file, int line));
void odeleted _((char* file, int line));
int  ocheck_pointer _((void *p, char* file, int line, char* msg));
int  ocheck_heap _((char* file, int line));
void oallog_begin _((const_char* logname, char* file, int line));
void oallog_end _((char* file, int line));

/*  local functions  */
static int  ocheck_block _((Block *p, char* file, int line, char* msg));
static void orestore_heap _((Block *corrupted));
static void oallog_printf _((
    const char *file,
    const int  line,
    const char *fmt,		/*  -> see printf(3) for the meaning  */
    ...
));





/* --------   G L 0 B A L   V A R I A B L E S   ----------------------------- */

static FILE *fp_log = NULL;







/*
 *  NAME
 *	mnew - malloc replacement
 *	cnew - calloc replacement
 *	renew - realloc replacement
 *
 *  DESCRIPTION
 *	mnrew() allocates the required memory and puts some data around the
 *	allocated block for administation and checking.
 *
 *	cnew() allocates the required memory and clears the reserved memory
 *	with 0's.
 *
 *	renew() allocates a new memory and returns the old content to the
 *	new location.
 *
 *  SEE ALSO
 *	malloc(3), calloc(3), realloc(3)
 *
 *  BUGS
 *	cnew() is not very efficient because it calls mnew() and clears then
 *	allocated memory with 0.
 *
 *	realloc() checks two times the given pointer: first before reallocation,
 *	then after reallocation, when it frees the old area (by calling
 *	'odelete()').
 *
 *	All time new memory is required also the heap is checked. This may
 *	slow down the application.
 */



void *mnew HAS_3_PARAMS (size_t, size, char*, file, int, line)
{
    Block *p;
    
    DBUG_ENTER ("mnew");

    /*  check heap first  */
    CHECK_HEAP (file, line);

    /*  check size  */
    if (size <= 0) {
	oallog_printf (
	    file,
	    line,
	    "WARNING: %ld bytes ordered!\n",
	    (long) size
	);
	if ((p = (Block *) malloc(size)) == NULL) {
	    DBUG_RETURN ((void*) NULL);
	}
	/*  in the other case the compiler seems to give back a pointer for   */
	/*  malloc(0) -> handle it...					      */
    }
    
    /*  allocate memory  */
    p = (Block *) malloc (sizeof(Block) + size + sizeof (end_magic));
    if (p == NULL) {
	DBUG_RETURN ((void*) NULL);
    }
    
    /*  mark begin and end of allocated memory  */
    p->size	   = size;
    p->magic	   = beg_magic;					    /* (4) b) */
    p->p_block_end = (((char *) p) + sizeof (Block) + size);
    strncpy (p->p_block_end, end_magic, sizeof (end_magic));	    /* (2) */
    memset ((void*) (p+1), 0xC5, size);				    /* (5) */

    /*  store the source name and line number  */
    p->file = file;
    p->line = line;
    
    /*  append it to the memory list  */
    p->next = NULL;
    if (p_LastBlock == NULL) {		/* it's the only block in the list */
	p->prev = NULL;
    	p_LastBlock  = p;
    	p_FirstBlock = p;
    } else {				/* append it at the end of the list */
        p->prev = p_LastBlock;
    	p_LastBlock->next = p;
    	p_LastBlock = p;
    }
    BlockCounter.allocated++;

    /*  return the allocated memory after the block  */
    DBUG_RETURN (p+1);
}



void *cnew HAS_4_PARAMS (size_t, nobj, size_t, size, char*, file, int, line)
{
    size_t bytes = nobj * size;
    void *p;
    
    DBUG_ENTER ("cnew");
    
    /*  check on arithmetik overflow  */
    if (bytes / nobj != size) {
    	oallog_printf (
    	    file,
    	    line,
    	    "arithmetik overflow in 'calloc (%lu, %lu)'!\n",
    	    (unsigned long) nobj,
    	    (unsigned long) size
    	);
    	DBUG_RETURN ((void*) NULL);
    }
    
    /*  allocate memory  */
    p = mnew (bytes, file, line);
    if (p) {
	memset (p, 0, bytes);
    }
    DBUG_RETURN (p);
}



void *renew HAS_4_PARAMS (void*, p, size_t, size, char*, file, int, line)
{
    Block   *b;
    void    *p_new;
    
    DBUG_ENTER ("renew");

    /*  check the given block  */
    if (p == NULL) {
	oallog_printf (file, line, "Can't reallocate NULL pointer!\n");
	DBUG_RETURN ((void*) NULL);
    }
    /*  check if the pointer is corrupted  */
    if (ocheck_pointer (p,file,line," to reallocate") == ERR_INVALID_POINTER) {
	DBUG_RETURN ((void*) NULL);
    }
    
    /*  allocate a new piece of memory (7)  */
    p_new = mnew (size, file, line);
    if (p_new == NULL) {
    	DBUG_RETURN ((void*) NULL);
    }
    
    /*  copy old content to new location  */
    b = ((Block *) p) - 1;
    memcpy (p_new, p, MIN (size, b->size));
    
    /*  free and shredder old block (7)  */
    odelete (p, file, line);
    
    DBUG_RETURN (p_new);
}





/*
 *  NAME
 *	odelete - free replacement
 *
 *  DESCRIPTION
 *	Checks the given pointer. If it is ok it frees the area and clears
 *	the area with garbage.
 *
 *  SEE ALSO
 *	free(3)
 *
 *  BUGS
 *	At each call also the heap is checked. This may slow down the
 *	application.
 */



void odelete HAS_3_PARAMS (void*, p, char*, file, int, line)
{
    Block *b, *prev, *next;
    
    DBUG_ENTER ("odelete");

    /*  check on NULL pointer  */
    if (p == NULL) {
	oallog_printf (file, line, "NULL pointer to delete!\n");
	DBUG_VOID_RETURN;
    }

    /*  check if the pointer is corrupted  */
    if (ocheck_pointer (p, file, line, " to delete") == ERR_INVALID_POINTER) {
	DBUG_VOID_RETURN;
    }

    /*  remove it from the memory list  */
    b = ((Block *) p) - 1;
    prev = b->prev;
    next = b->next;
    if (prev == NULL) {
    	/* we have to delete the FIRST block */
    	p_FirstBlock = next;
    } else {
    	prev->next = next;
    }
    if (next == NULL) {
    	/* we have to delete the LAST block */
    	p_LastBlock = prev;
    } else {
    	next->prev = prev;
    }

    /*  and now shredder the content and free it (6)  */
    memset ((void*) b, 0x3A, b->size + sizeof (Block));
    BlockCounter.freed++;
    free ((void *) b);
    
    /*  check the heap  */
    CHECK_HEAP (file, line);
    DBUG_VOID_RETURN;
}





/*
 *  NAME
 *	odeleted
 *
 *  DESCRIPTION
 *	Checks if all memory is freed. If not it prints a message and checks
 *	the remaining blocks.
 */



void odeleted HAS_2_PARAMS (char*, file, int, line)
{
    Block *b;

    DBUG_ENTER ("odeleted");

    /*  check if there are unfreed blocks remaining  */
    if (BlockCounter.allocated != BlockCounter.freed) {
	oallog_printf (
	    file,
	    line,
	    "%d blocks reserved, but only %d freed!\n",
	    BlockCounter.allocated,
	    BlockCounter.freed
	);
    }
    for (b = p_FirstBlock; b != NULL; b = b->next) {
    	oallog_printf (
    	    file,
    	    line,
    	    "%lu bytes reserved at %s(%d) not freed!\n",
    	    (unsigned long) b->size,
    	    b->file,
    	    b->line
    	);
    }

    /*  check the remaining heap  */
    CHECK_HEAP (file, line);
    oallog_end (file, line);
    DBUG_VOID_RETURN;
}





/*
 *  NAME
 *	oallog_begin
 *	oallog_end
 *
 *  DESCRIPTION
 *	With oallog_begin() the given log file is opened. Also an already
 *	opened file is closed before.
 *
 *	oallog_end() closes the given log file. It is automatically called
 *	by odeleted().
 */



void oallog_begin HAS_3_PARAMS (
    const_char*,    logname,
    char*, 	    file,
    int,	    line
)
{
    DBUG_ENTER ("oallog_begin");
    
    /*  close old log file (if there is one)  */
    oallog_end (file, line);
    
    /*  open file (if <logname> given)  */
    if (logname == NULL) {
    	fp_log = stderr;
    	DBUG_VOID_RETURN;
    }
    fp_log = fopen (logname, "w");
    if (fp_log == NULL) {
    	perror (logname);
    } else {
	time_t t = time(NULL);

    	oallog_printf (
    	    file,
    	    line,
    	    "log-file opened at %s",
    	    ctime (&t)
    	);
    }

    DBUG_VOID_RETURN;
}



void oallog_end HAS_2_PARAMS (
    char*,  file,
    int,    line
)
{
    DBUG_ENTER ("oallog_end");

    /*  close old log file (if there is one)  */
    if ((fp_log != stderr) && (fp_log != NULL)) {
	time_t t = time(NULL);

    	oallog_printf (
    	    file,
    	    line,
    	    "log-file closed at %s",
    	    ctime (&t)
    	);
    	if (ferror(fp_log)) {
    	    perror ("log-file");
    	}
    	fclose (fp_log);
    	fp_log = stderr;
    }
    
    DBUG_VOID_RETURN;
}







/* --------   S U P P O R T E D   S T R I N G   F U N C T I O N S   --------- */





/*
 *  ostrdup
 *
 *  Duplicates a string and saves it into a saved memory allocation.
 *  (see strsave() in libostring.c)
 */

char *ostrdup HAS_3_PARAMS (
    char*, srcstr,		    /* source string */
    char*, file,
    int,   line
)
{
    char *deststr;		    /* destination string */
    int l_deststr;		    /* length of the destination string */

    DBUG_ENTER ("ostrdup");
    if (srcstr == NULL) {
    	oallog_printf (
    	    file,
    	    line,
    	    "WARNING: strdup() called with NULL pointer!\n"
    	);
    	DBUG_RETURN ((void*) NULL);
    }
    l_deststr = strlen (srcstr) + 1;
    if ((deststr = (char *) mnew ((unsigned) l_deststr, file, line)) != NULL) {
	strcpy (deststr, srcstr);
    }
    DBUG_RETURN (deststr);
}







/* --------   A D D I T I O N A L   F U N C T I O N S   --------------------- */




/*
 *  NAME
 *	ocheck_block
 *	ocheck_pointer
 *	ocheck_heap
 *	orestore_heap
 *
 *  DESCRIPTION
 *	ocheck_block() checks if the given block is not corrupted.
 *
 *	ocheck_pointer() checks if the given pointer is a valid pointer.
 *	If the end of the allocated block is overwritten the end mark is
 *	written back to avoid error messages from later checks and to
 *	recognize further overwriting of the memory end.
 *
 *	ocheck_heap() checks all the pointers in the memory list.
 *
 *	orestore_heap() searches the corrupted block from the beginning and
 *	from the end and makes a new list without the corrupted block(s).
 *	Next each header is marked as uncorrupted.
 *
 *  RETURN_VALUE
 *	ocheck_pointer:
 *	    NO_ERROR
 *	    ERR_INVALID_POINTER
 *	    ERR_MEMORY_OVERWRITTEN
 *
 *  BUGS
 *	The given pointer for ocheck_pointer() must be a pointer to the
 *	beginning of the allocated memory.
 */



static int ocheck_block HAS_4_PARAMS (
    Block*, b,
    char*,  file,
    int,    line,
    char*,  more_message			/* additional error message */
)
{
    DBUG_ENTER ("ocheck_block");
    assert (b);
    
    /*  check the block beginning  */
    if (!VALID_BLOCK_POINTER (b) || (b->magic != beg_magic)) {
	oallog_printf (file, line, "Invalid pointer");
	if (more_message) {
	    (void) fputs (more_message, fp_log);
	}
	(void) fputs ("!\n", fp_log);
	DBUG_RETURN (ERR_INVALID_POINTER);
    }
    
    /*  check if block end was overwritten  */
    if (strncmp (b->p_block_end, end_magic, sizeof (end_magic)) != 0) {
	oallog_printf (
	    file,
	    line,
	    "End of %lu-byte-block - reserved at %s(%d) - "
	    "overwritten!\n",
	    (unsigned long) b->size,
	    b->file,
	    b->line
	);
	strncpy (b->p_block_end, end_magic, sizeof (end_magic));
	DBUG_RETURN (ERR_MEMORY_OVERWRITTEN);
    }
    
    /*  block seems to be ok  */
    DBUG_RETURN (NO_ERROR);
}



int ocheck_pointer HAS_4_PARAMS (
    void*, p,
    char*, file,
    int,   line,
    char*,  more_message			/* additional error message */
)
{
    Block *b;
    
    DBUG_ENTER ("ocheck_pointer");
    assert (p);

    b = ((Block *) p) - 1;
    DBUG_RETURN (ocheck_block (b, file, line, more_message));
}



int ocheck_heap HAS_2_PARAMS (char*, file, int, line)
{
    register Block  *b;
    
    DBUG_ENTER ("ocheck_heap");

    /*  check the heap  */
    for (b = p_FirstBlock; b != NULL; b = b->next) {
	if (ocheck_block (b, file, line, NULL) == ERR_INVALID_POINTER) {
	    oallog_printf (
	    	file,
	    	line,
	    	"The heap is corrupted - trying to restore...\n"
	    );
	    orestore_heap (b);
	    DBUG_RETURN (HEAP_CORRUPTED);
	    break;
	}
    }
    
    /*  heap seems to be ok  */
    DBUG_RETURN (HEAP_OK);
}



static void orestore_heap HAS_1_PARAM (Block*, corrupted)
{
    Block *b;
    Block *p_LastBlock1_ok  = NULL;
    Block *p_FirstBlock2_ok = NULL;
    
    DBUG_ENTER ("orestore_heap");

    /*  go to corrupted block from the beginning  */
    for (b = p_FirstBlock; b != NULL; b = b->next) {
    	if (!VALID_BLOCK_POINTER (b) || (b->prev != p_LastBlock1_ok)
    				     || (b == corrupted)) {
    	    break;
    	}
    	p_LastBlock1_ok = b;
    }
    
    /*  go to corrupted block from the end  */
    for (b = p_LastBlock; b != NULL; b = b->prev) {
    	if (!VALID_BLOCK_POINTER (b) || (b->next != p_FirstBlock2_ok)
    				     || (b == corrupted)) {
    	    break;
    	}
    	p_FirstBlock2_ok = b;
    }
    
    /*  now merge the two list parts  */
    if (p_LastBlock1_ok && p_FirstBlock2_ok) {
    	p_LastBlock1_ok->next  = p_FirstBlock2_ok;
    	p_FirstBlock2_ok->prev = p_LastBlock1_ok;
    }
    
    /*  restore each header (mark it as uncorrupted)  */
    for (b = p_FirstBlock; b != NULL; b = b->next) {
	b->magic = beg_magic;
    }
    
    DBUG_VOID_RETURN;
}





/*
 *  NAME
 *	oallog_printf
 *
 *  DESCRIPTION
 *	Prints a message to the log file. The format of the message is:
 *
 *	    <file>(<line>): ...
 *
 *	oallog_end() closes the given log file. It is automatically called
 *	by odeleted().
 */



#ifndef _NO_PROTO
static void oallog_printf (
    const char *file,
    const int  line,
    const char *fmt,		/*  -> see printf(3) for the meaning  */
    ...
)
#else
/*VARARGS*/
static void oallog_printf (va_alist) va_dcl
#endif
{
#ifdef _NO_PROTO
    char *fmt;
    char *file;
    int  line;
#endif
    va_list ap;

    DBUG_ENTER ("oallog_printf");
    
#ifndef _NO_PROTO
    va_start (ap, fmt);
#else
    va_start (ap);
    fmt  = (char *) va_arg (ap, char *);
    file = (char *) va_arg (ap, char *);
    line = (int) va_arg (ap, int);
#endif

    /*  print the message to the log file  */
    (void) fprintf  (fp_log, "%s(%d): ", file, line);
    (void) vfprintf (fp_log, fmt, ap);
    fflush (fp_log);

    DBUG_VOID_RETURN;
}
